#include "mpdimpl.h"

bool GenAuthenticationStrings(char *append, char *crypted)
{
    int stamp;
    char *crypted_internal;
    char phrase[MPD_PASSPHRASE_MAX_LENGTH+1];
    char phrase_internal[MPD_PASSPHRASE_MAX_LENGTH+1];

    srand(GetTickCount());
    stamp = rand();

    if (!ReadMPDRegistry("phrase", phrase))
	return false;

    _snprintf(phrase_internal, MPD_PASSPHRASE_MAX_LENGTH, "%s%d", phrase, stamp);
    sprintf(append, "%d", stamp);

    //dbg_printf("GenAuthenticationStrings: calling crypt on '%s'\n", phrase_internal);
    crypted_internal = crypt(phrase_internal, MPD_SALT_VALUE);
    if (strlen(crypted_internal) > MPD_PASSPHRASE_MAX_LENGTH)
	return false;
    strcpy(crypted, crypted_internal);

    memset(phrase, 0, MPD_PASSPHRASE_MAX_LENGTH);
    memset(phrase_internal, 0, MPD_PASSPHRASE_MAX_LENGTH);

    return true;
}

bool AuthenticateConnection(MPD_Context *p)
{
    int ret_val;

    // generate the challenge string and the encrypted result
    if (!GenAuthenticationStrings(p->pszOut, p->pszCrypt))
    {
	err_printf("AuthenticateConnection: failed to generate the authentication strings\n");
	RemoveContext(p);
	return false;
    }

    // write the challenge string
    if (WriteString(p->bfd, p->pszOut) == SOCKET_ERROR)
    {
	err_printf("AuthenticateConnection: Writing challenge string failed, error %d\n", WSAGetLastError());
	RemoveContext(p);
	return false;
    }

    // read the response
    if (!ReadString(p->bfd, p->pszIn))
    {
	err_printf("AuthenticateConnection: Reading challenge response failed, error %d\n", WSAGetLastError());
	RemoveContext(p);
	return false;
    }

    // compare the response with the encrypted result and write success or failure
    if (strcmp(p->pszIn, p->pszCrypt) == 0)
	ret_val = WriteString(p->bfd, "SUCCESS");
    else
	ret_val = WriteString(p->bfd, "FAIL");
    if (ret_val == SOCKET_ERROR)
    {
	err_printf("AuthenticateConnection: Writing authentication result failed, error %d\n", WSAGetLastError());
	RemoveContext(p);
	return false;
    }

    // read the type of connection
    if (!ReadString(p->bfd, p->pszIn))
    {
	err_printf("AuthenticateConnection: Reading the connection type failed, error %d\n", WSAGetLastError());
	RemoveContext(p);
	return false;
    }

    // set the state appropriate for the type of connection
    if (stricmp(p->pszIn, "console") == 0)
    {
	dbg_printf("AuthenticateConnection: bfd[%s] nType: MPD_CONSOLE_SOCKET\n", bto_string(p->bfd));
	p->nType = MPD_CONSOLE_SOCKET;
	p->nLLState= MPD_READING_CMD;
    }
    else if (strnicmp(p->pszIn, "left ", 5) == 0)
    {
	dbg_printf("AuthenticateConnection: bfd[%s] nType: MPD_LEFT_SOCKET\n", bto_string(p->bfd));
	p->nType = MPD_LEFT_SOCKET;
	p->nLLState= MPD_READING_CMD;
	strncpy(p->pszHost, &p->pszIn[5], MAX_HOST_LENGTH);
	p->pszHost[MAX_HOST_LENGTH-1] = '\0';
    }
    else if (strnicmp(p->pszIn, "right ", 6) == 0)
    {
	dbg_printf("AuthenticateConnection: bfd[%s] nType: MPD_RIGHT_SOCKET\n", bto_string(p->bfd));
	p->nType = MPD_RIGHT_SOCKET;
	p->nLLState= MPD_READING_CMD;
	strncpy(p->pszHost, &p->pszIn[6], MAX_HOST_LENGTH);
	p->pszHost[MAX_HOST_LENGTH-1] = '\0';
    }
    else
    {
	err_printf("AuthenticateConnection: unknown socket type read: '%s'\n", p->pszIn);
	RemoveContext(p);
	return false;
    }
    p->nState = MPD_IDLE;

    return true;
}
