#include <stdio.h>
#include <stdlib.h>
#include "LaunchProcess.h"
#include "global.h"
#include "bsocket.h"
#include "mpirun.h"
#include "mpdutil.h"

struct ProcessWaitThreadArg
{
    int n;
    int *pSocket;
    int *pId;
    int *pRank;
    int bfdAbort;
};

struct ProcessWaitAbortThreadArg
{
    int bfdAbort;
    int bfdStop;
    int n;
    int *pSocket;
};

// Function name	: ProcessWaitAbort
// Description	    : 
// Return type		: void 
// Argument         : ProcessWaitAbortThreadArg *pArg
void ProcessWaitAbort(ProcessWaitAbortThreadArg *pArg)
{
    int n, i;
    bfd_set readset;

    BFD_ZERO(&readset);
    BFD_SET(pArg->bfdAbort, &readset);
    BFD_SET(pArg->bfdStop, &readset);

    n = bselect(0, &readset, NULL, NULL, NULL);

    if (n == SOCKET_ERROR)
    {
	PrintError(WSAGetLastError(), "bselect failed\n");fflush(stdout);
	for (i=0; i<pArg->n; i++)
	{
	    beasy_closesocket(pArg->pSocket[i]);
	}
	beasy_closesocket(pArg->bfdAbort);
	beasy_closesocket(pArg->bfdStop);
	return;
    }
    if (n == 0)
    {
	printf("ProcessWaitAbort: bselect returned zero sockets available\n");fflush(stdout);
	for (i=0; i<pArg->n; i++)
	{
	    beasy_closesocket(pArg->pSocket[i]);
	}
	beasy_closesocket(pArg->bfdAbort);
	beasy_closesocket(pArg->bfdStop);
	return;
    }
    if (BFD_ISSET(pArg->bfdAbort, &readset))
    {
	for (i=0; i<pArg->n; i++)
	{
	    beasy_send(pArg->pSocket[i], "x", 1);
	}
    }
    for (i=0; i<pArg->n; i++)
    {
	beasy_closesocket(pArg->pSocket[i]);
    }
    beasy_closesocket(pArg->bfdAbort);
    beasy_closesocket(pArg->bfdStop);
}

// Function name	: UnmapDrives
// Description	    : 
// Return type		: bool 
// Argument         : int bfd
bool UnmapDrives(int bfd)
{
    char pszStr[256];
    if (g_pDriveMapList && !g_bNoDriveMapping)
    {
	MapDriveNode *pNode = g_pDriveMapList;
	while (pNode)
	{
	    sprintf(pszStr, "unmap drive=%c", pNode->cDrive);
	    if (WriteString(bfd, pszStr) == SOCKET_ERROR)
	    {
		PrintError(WSAGetLastError(), "ERROR: Unable to send unmap command\n");
		beasy_closesocket(bfd);
		SetEvent(g_hAbortEvent);
		return false;
	    }
	    if (!ReadString(bfd, pszStr))
	    {
		PrintError(WSAGetLastError(), "ERROR: Unable to read the result of unmap command\n");
		beasy_closesocket(bfd);
		SetEvent(g_hAbortEvent);
		return false;
	    }
	    if (stricmp(pszStr, "SUCCESS"))
	    {
		printf("ERROR: Unable to unmap %c: %s\r\n%s", pNode->cDrive, pNode->pszShare, pszStr);
		beasy_closesocket(bfd);
		SetEvent(g_hAbortEvent);
		return false;
	    }
	    pNode = pNode->pNext;
	}
    }
    return true;
}

// Function name	: ProcessWait
// Description	    : 
// Return type		: void 
// Argument         : ProcessWaitThreadArg *pArg
void ProcessWait(ProcessWaitThreadArg *pArg)
{
    int i, j, n;
    bfd_set totalset, readset;
    char str[256];
    
    BFD_ZERO(&totalset);
    
    BFD_SET(pArg->bfdAbort, &totalset);
    for (i=0; i<pArg->n; i++)
    {
	BFD_SET(pArg->pSocket[i], &totalset);
    }
    
    while (pArg->n)
    {
	readset = totalset;
	n = bselect(0, &readset, NULL, NULL, NULL);
	if (n == SOCKET_ERROR)
	{
	    PrintError(WSAGetLastError(), "bselect failed\n");fflush(stdout);
	    for (i=0, j=0; i<pArg->n; i++, j++)
	    {
		while (pArg->pSocket[j] == BFD_INVALID_SOCKET)
		    j++;
		beasy_closesocket(pArg->pSocket[j]);
		pArg->pSocket[j] = BFD_INVALID_SOCKET;
	    }
	    return;
	}
	if (n == 0)
	{
	    printf("ProcessWait: bselect returned zero sockets available");fflush(stdout);
	    for (i=0, j=0; i<pArg->n; i++, j++)
	    {
		while (pArg->pSocket[j] == BFD_INVALID_SOCKET)
		    j++;
		beasy_closesocket(pArg->pSocket[j]);
		pArg->pSocket[j] = BFD_INVALID_SOCKET;
	    }
	    return;
	}

	if (BFD_ISSET(pArg->bfdAbort, &readset))
	{
	    for (i=0; pArg->n > 0; i++)
	    {
		while (pArg->pSocket[i] == BFD_INVALID_SOCKET)
		    i++;
		sprintf(str, "kill %d", pArg->pId[i]);
		//printf("%d:%s\n", __LINE__, str);fflush(stdout);
		WriteString(pArg->pSocket[i], str);

		int nRank = pArg->pRank[i];
		if (g_nNproc > FORWARD_NPROC_THRESHOLD)
		{
		    if (nRank > 0 && (g_nNproc/2) > nRank)
		    {
			//printf("rank %d(%d) stopping forwarder\n", nRank, g_pProcessLaunchId[i]);fflush(stdout);
			sprintf(str, "stopforwarder port=%d abort=yes", g_pForwardHost[nRank].nPort);
			WriteString(pArg->pSocket[i], str);
		    }
		}

		// Is this taken care of somewhere else?
		//UnmapDrives(pArg->pSocket[i], g_pDriveMapList);

		sprintf(str, "freeprocess %d", pArg->pId[i]);
		WriteString(pArg->pSocket[i], str);
		ReadString(pArg->pSocket[i], str);
		WriteString(pArg->pSocket[i], "done");
		beasy_closesocket(pArg->pSocket[i]);
		pArg->pSocket[i] = BFD_INVALID_SOCKET;
		pArg->n--;
	    }
	    return;
	}
	for (i=0; n>0; i++)
	{
	    while (pArg->pSocket[i] == BFD_INVALID_SOCKET)
		i++;
	    if (BFD_ISSET(pArg->pSocket[i], &readset))
	    {
		if (ReadString(pArg->pSocket[i], str))
		{
		    int nRank = pArg->pRank[i];
		    
		    if (strnicmp(str, "FAIL", 4) == 0)
		    {
			// get the error
			sprintf(str, "geterror %d", pArg->pId[i]);
			WriteString(pArg->pSocket[i], str);
			ReadString(pArg->pSocket[i], str);
			printf("getexitcode(rank %d) failed: %s\n", nRank, str);fflush(stdout);
			if (g_bUseJobHost)
			{
			    UpdateJobKeyValue(nRank, "error", str);
			    
			    // get the time the process exited
			    sprintf(str, "getexittime %d", pArg->pId[i]);
			    WriteString(pArg->pSocket[i], str);
			    ReadString(pArg->pSocket[i], str);
			    UpdateJobKeyValue(nRank, "exittime", str);
			}
			
			if (beasy_send(g_bfdBreak, "x", 1) == SOCKET_ERROR)
			{
			    printf("Hard abort.\n");fflush(stdout);
			    ExitProcess(-1);
			}
		    }
		    else
		    {
			if (g_bUseJobHost)
			{
			    strtok(str, ":"); // strip the extra data from the string
			    UpdateJobKeyValue(nRank, "exitcode", str);
			    
			    // get the time the process exited
			    sprintf(str, "getexittime %d", pArg->pId[i]);
			    WriteString(pArg->pSocket[i], str);
			    ReadString(pArg->pSocket[i], str);
			    UpdateJobKeyValue(nRank, "exittime", str);
			}

			if (!g_bNoMPI)
			{
			    sprintf(str, "getmpifinalized %d", pArg->pId[i]);
			    WriteString(pArg->pSocket[i], str);
			    ReadString(pArg->pSocket[i], str);
			    if (stricmp(str, "yes") != 0)
			    {
				if (stricmp(str, "no") != 0)
				    printf("getmpifinalized returned: %s\n", str);
				else
				{
				    if (!g_bSuppressErrorOutput)
					printf("process exited without calling MPIFinalize\n");
				}
				fflush(stdout);
				beasy_send(g_bfdBreak, "x", 1);
			    }
			}
		    }
		    
		    if (g_nNproc > FORWARD_NPROC_THRESHOLD)
		    {
			if (nRank > 0 && (g_nNproc/2) > nRank)
			{
			    sprintf(str, "stopforwarder port=%d abort=no", g_pForwardHost[nRank].nPort);
			    WriteString(pArg->pSocket[i], str);
			}
		    }
		    
		    UnmapDrives(pArg->pSocket[i]);
		    
		    sprintf(str, "freeprocess %d", pArg->pId[i]);
		    WriteString(pArg->pSocket[i], str);
		    ReadString(pArg->pSocket[i], str);
		    
		    WriteString(pArg->pSocket[i], "done");
		    beasy_closesocket(pArg->pSocket[i]);
		    BFD_CLR(pArg->pSocket[i], &totalset);
		    pArg->pSocket[i] = BFD_INVALID_SOCKET;
		    n--;
		    pArg->n--;
		}
		else
		{
		    PrintError(WSAGetLastError(), "ProcessWait:Reading the exit code for process %d failed\n", i);fflush(stdout);
		    beasy_closesocket(pArg->pSocket[i]);
		    BFD_CLR(pArg->pSocket[i], &totalset);
		    pArg->pSocket[i] = BFD_INVALID_SOCKET;
		    n--;
		    pArg->n--;
		    if (beasy_send(g_bfdBreak, "x", 1) == SOCKET_ERROR)
		    {
			printf("Unable to abort processes.\n");fflush(stdout);
			ExitProcess(-1);
		    }
		    //return;
		}
	    }
	}
    }
}

// Function name	: WaitForExitCommands
// Description	    : 
// Return type		: void 
void WaitForExitCommands()
{
    bool bKillSent = false;
    if (g_nNumProcessSockets < FD_SETSIZE)
    {
	int i, n;
	bfd_set totalset, readset;
	char str[256];
	int break_bfd;

	//printf("waiting for %d processes\n", g_nNumProcessSockets);fflush(stdout);
	/*
	for (i=0; i<g_nNumProcessSockets; i++)
	{
	    printf("socket[%d] = id %d\n", g_pProcessSocket[i], g_pProcessLaunchId[i]);
	}
	fflush(stdout);
	*/

	MakeLoop(&break_bfd, &g_bfdBreak);
	SetEvent(g_hBreakReadyEvent); // allow a break to happen if the user has already hit Ctrl-C

	BFD_ZERO(&totalset);

	BFD_SET(break_bfd, &totalset);
	for (i=0; i<g_nNumProcessSockets; i++)
	{
	    BFD_SET(g_pProcessSocket[i], &totalset);
	}

	while (g_nNumProcessSockets)
	{
	    readset = totalset;
	    n = bselect(0, &readset, NULL, NULL, NULL);
	    if (n == SOCKET_ERROR)
	    {
		PrintError(WSAGetLastError(), "WaitForExitCommands: bselect failed\n");fflush(stdout);
		for (i=0; g_nNumProcessSockets > 0; i++)
		{
		    while (g_pProcessSocket[i] == BFD_INVALID_SOCKET)
			i++;
		    //printf("closing socket [%d]: %s\n", i, bto_string(g_pProcessSocket[i]));fflush(stdout);
		    beasy_closesocket(g_pProcessSocket[i]);
		    g_pProcessSocket[i] = BFD_INVALID_SOCKET;
		    g_nNumProcessSockets--;
		}
		return;
	    }
	    if (n == 0)
	    {
		printf("WaitForExitCommands: bselect returned zero sockets available\n");fflush(stdout);
		for (i=0; g_nNumProcessSockets > 0; i++)
		{
		    while (g_pProcessSocket[i] == BFD_INVALID_SOCKET)
			i++;
		    //printf("closing socket [%d]: %s\n", i, bto_string(g_pProcessSocket[i]));fflush(stdout);
		    beasy_closesocket(g_pProcessSocket[i]);
		    g_pProcessSocket[i] = BFD_INVALID_SOCKET;
		    g_nNumProcessSockets--;
		}
		return;
	    }
	    else
	    {
		if (BFD_ISSET(break_bfd, &readset))
		{
		    int num_read = beasy_receive(break_bfd, str, 1);
		    if (num_read == 0 || num_read == SOCKET_ERROR)
		    {
			BFD_CLR(break_bfd, &totalset);
		    }
		    else
		    {
			if (!bKillSent)
			{
			    printf("Sending kill commands to launched processes\n");fflush(stdout);
			    for (int j=0, i=0; i<g_nNumProcessSockets; i++, j++)
			    {
				while (g_pProcessSocket[j] == BFD_INVALID_SOCKET)
				    j++;
				sprintf(str, "kill %d", g_pProcessLaunchId[j]);
				//printf("%d:%s\n", __LINE__, str);fflush(stdout);
				//printf("kill %d (id[%d])\n", g_pProcessLaunchId[j], j);fflush(stdout);
				if (WriteString(g_pProcessSocket[j], str) == SOCKET_ERROR)
				{
				    printf("writing kill command failed\n");fflush(stdout);
				}
			    }
			    bKillSent = true;
			}
		    }
		    n--;
		}
		for (i=0; n>0; i++)
		{
		    while (g_pProcessSocket[i] == BFD_INVALID_SOCKET)
			i++;
		    if (BFD_ISSET(g_pProcessSocket[i], &readset))
		    {
			if (ReadString(g_pProcessSocket[i], str))
			{
			    int nRank = g_pLaunchIdToRank[i];
			    
			    if (strnicmp(str, "FAIL", 4) == 0)
			    {
				sprintf(str, "geterror %d", g_pProcessLaunchId[i]);
				WriteString(g_pProcessSocket[i], str);
				ReadString(g_pProcessSocket[i], str);
				printf("getexitcode(rank %d) failed: %s\n", nRank, str);fflush(stdout);
				UpdateJobKeyValue(nRank, "error", str);
				
				if (g_bUseJobHost)
				{
				    // get the time the process exited
				    sprintf(str, "getexittime %d", g_pProcessLaunchId[i]);
				    WriteString(g_pProcessSocket[i], str);
				    ReadString(g_pProcessSocket[i], str);
				    UpdateJobKeyValue(nRank, "exittime", str);
				}
				
				if (beasy_send(g_bfdBreak, "x", 1) == SOCKET_ERROR)
				{
				    printf("Aborting.\n");fflush(stdout);
				    ExitProcess(-1);
				}
			    }
			    else
			    {
				//printf("ExitProcess: %s\n", str);fflush(stdout);
				strtok(str, ":"); // strip the extra data from the string
				UpdateJobKeyValue(nRank, "exitcode", str);
				
				if (g_bUseJobHost)
				{
				    // get the time the process exited
				    sprintf(str, "getexittime %d", g_pProcessLaunchId[i]);
				    WriteString(g_pProcessSocket[i], str);
				    ReadString(g_pProcessSocket[i], str);
				    UpdateJobKeyValue(nRank, "exittime", str);
				}

				if (!g_bNoMPI)
				{
				    sprintf(str, "getmpifinalized %d", g_pProcessLaunchId[i]);
				    WriteString(g_pProcessSocket[i], str);
				    ReadString(g_pProcessSocket[i], str);
				    if (stricmp(str, "yes") != 0)
				    {
					if (stricmp(str, "no") != 0)
					    printf("getmpifinalized returned: %s\n", str);
					else
					{
					    if (!g_bSuppressErrorOutput)
						printf("process exited without calling MPIFinalize\n");
					}
					fflush(stdout);
					beasy_send(g_bfdBreak, "x", 1);
				    }
				}
			    }
			    
			    if (g_nNproc > FORWARD_NPROC_THRESHOLD)
			    {
				if (nRank > 0 && (g_nNproc/2) > nRank)
				{
				    //printf("rank %d(%d) stopping forwarder\n", nRank, g_pProcessLaunchId[i]);fflush(stdout);
				    sprintf(str, "stopforwarder port=%d abort=no", g_pForwardHost[nRank].nPort);
				    WriteString(g_pProcessSocket[i], str);
				}
			    }

			    UnmapDrives(g_pProcessSocket[i]);
			    
			    sprintf(str, "freeprocess %d", g_pProcessLaunchId[i]);
			    WriteString(g_pProcessSocket[i], str);
			    ReadString(g_pProcessSocket[i], str);
			    
			    WriteString(g_pProcessSocket[i], "done");
			    //printf("closing socket [%d]: %s\n", i, bto_string(g_pProcessSocket[i]));fflush(stdout);
			    BFD_CLR(g_pProcessSocket[i], &totalset);
			    beasy_closesocket(g_pProcessSocket[i]);
			    g_pProcessSocket[i] = BFD_INVALID_SOCKET;
			    //printf("closing socket [%d]\n", i);fflush(stdout);
			    n--;
			    g_nNumProcessSockets--;
			    //printf("(E:%d)", g_pProcessLaunchId[i]);fflush(stdout);
			}
			else
			{
			    if (WSAGetLastError() != 0)
				PrintError(WSAGetLastError(), "WaitForExitCommands:Reading the exit code for process %d failed.\n", i);
			    else
			    {
				printf("WaitForExitCommands:Reading the exit code for process %d failed.\n", i);
				fflush(stdout);
			    }
			    BFD_CLR(g_pProcessSocket[i], &totalset);
			    sprintf(str, "kill %d", g_pProcessLaunchId[i]);
			    WriteString(g_pProcessSocket[i], str);
			    WriteString(g_pProcessSocket[i], "done");
			    //printf("closing socket [%d]: %s\n", i, bto_string(g_pProcessSocket[i]));fflush(stdout);
			    beasy_closesocket(g_pProcessSocket[i]);
			    g_pProcessSocket[i] = BFD_INVALID_SOCKET;
			    //printf("closing socket [%d]\n", i);fflush(stdout);
			    n--;
			    g_nNumProcessSockets--;
			    if (beasy_send(g_bfdBreak, "x", 1) == SOCKET_ERROR)
			    {
				printf("Unable to abort processes.\n");fflush(stdout);
				ExitProcess(-1);
			    }
			}
		    }
		}
	    }
	}
	
	beasy_closesocket(g_bfdBreak);
	g_bfdBreak = BFD_INVALID_SOCKET;
	delete g_pProcessSocket;
	delete g_pProcessLaunchId;
	delete g_pLaunchIdToRank;
	g_pProcessSocket = NULL;
	g_pProcessLaunchId = NULL;
	g_pLaunchIdToRank = NULL;
    }
    else
    {
	DWORD dwThreadID;
	int num = (g_nNumProcessSockets / (FD_SETSIZE-1)) + 1;
	HANDLE *hThread = new HANDLE[num];
	int *pAbortBFD = new int[num];
	int bfdStop;
	ProcessWaitThreadArg *arg = new ProcessWaitThreadArg[num];
	ProcessWaitAbortThreadArg *arg2 = new ProcessWaitAbortThreadArg;
        int i;
	for (i=0; i<num; i++)
	{
	    if (i == num-1)
		arg[i].n = g_nNumProcessSockets % (FD_SETSIZE-1);
	    else
		arg[i].n = (FD_SETSIZE-1);
	    arg[i].pSocket = &g_pProcessSocket[i*(FD_SETSIZE-1)];
	    arg[i].pId = &g_pProcessLaunchId[i*(FD_SETSIZE-1)];
	    arg[i].pRank = &g_pLaunchIdToRank[i*(FD_SETSIZE-1)];
	    MakeLoop(&arg[i].bfdAbort, &pAbortBFD[i]);
	}
	for (i=0; i<num; i++)
	{
	    hThread[i] = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ProcessWait, &arg[i], 0, &dwThreadID);
	}
	MakeLoop(&arg2->bfdAbort, &g_bfdBreak);
	MakeLoop(&arg2->bfdStop, &bfdStop);

	HANDLE hWaitAbortThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ProcessWaitAbort, arg2, 0, &dwThreadID);

	SetEvent(g_hBreakReadyEvent);

	WaitForMultipleObjects(num, hThread, TRUE, INFINITE);
	for (i=0; i<num; i++)
	    CloseHandle(hThread[i]);
	delete hThread;
	delete arg;

	beasy_send(bfdStop, "x", 1);
	beasy_closesocket(bfdStop);
	WaitForSingleObject(hWaitAbortThread, 10000);
	delete pAbortBFD;
	delete arg2;
	CloseHandle(hWaitAbortThread);

	beasy_closesocket(g_bfdBreak);
	g_bfdBreak = BFD_INVALID_SOCKET;
	delete g_pProcessSocket;
	delete g_pProcessLaunchId;
	delete g_pLaunchIdToRank;
	g_pProcessSocket = NULL;
	g_pProcessLaunchId = NULL;
	g_pLaunchIdToRank = NULL;
    }
    //printf("WaitForExitCommands returning\n");fflush(stdout);
}
